@@ -1,8 +1,10 @@
 
+from pyroute2.netlink.iproute import IPRSocket
 from pyroute2.netlink.iproute import IPRoute
 from pyroute2.netlink.ipdb import IPDB
 from pyroute2.netlink.taskstats import TaskStats
 
+make_pep8_happy = IPRSocket
 make_pep8_happy = IPRoute
 make_pep8_happy = IPDB
 make_pep8_happy = TaskStats
