@@ -324,12 +324,16 @@ class NetlinkSocket(socket.socket):
         socket.socket.__init__(self, socket.AF_NETLINK,
                                socket.SOCK_DGRAM, family)
         self.pid = os.getpid()
-        self.groups = None
+        self.groups = 0
+        self.marshal = None
 
     def bind(self, groups=0):
         self.groups = groups
         socket.socket.bind(self, (self.pid, self.groups))
 
+    def get(self):
+        return self.marshal.parse(self.recv(16384))
+
 
 class ssl_credentials(object):
     def __init__(self, key, cert, ca):
@@ -352,7 +356,7 @@ class masq_record(object):
 
 class IOThread(threading.Thread):
     def __init__(self):
-        threading.Thread.__init__(self)
+        threading.Thread.__init__(self, name='Netlink I/O')
         self.setDaemon(True)
         self.pid = os.getpid()
         self._nonce = 0
@@ -378,6 +382,7 @@ class IOThread(threading.Thread):
         self.broadcast = set()    # set(socket, socket...)
         self.ssl_keys = {}        # {url: ssl_credentials(), url:...}
         self.mirror = False
+        self.callbacks = []       # [(predicate, callback, args), ...]
         # secret; write non-zero byte as terminator
         self.secret = os.urandom(15)
         self.secret += '\xff'
@@ -385,15 +390,18 @@ class IOThread(threading.Thread):
         # control in-process communication only
         self.sctl, self.control = pairPipeSockets()
         self.add_client(self.sctl)
-        self._sctl_thread = threading.Thread(target=self._sctl)
+        self._sctl_thread = threading.Thread(target=self._sctl,
+                                             name='IPC init')
         self._sctl_thread.start()
         # masquerade cache expiration
-        self._expire_thread = threading.Thread(target=self._expire_masq)
+        self._expire_thread = threading.Thread(target=self._expire_masq,
+                                               name='Masquerade cache')
         self._expire_thread.setDaemon(True)
         self._expire_thread.start()
         # buffers reassembling
         self.buffers = Queue.Queue()
-        self._feed_thread = threading.Thread(target=self._feed_buffers)
+        self._feed_thread = threading.Thread(target=self._feed_buffers,
+                                             name='Reasm and parsing')
         self._feed_thread.setDaemon(True)
         self._feed_thread.start()
         # debug
@@ -422,7 +430,7 @@ class IOThread(threading.Thread):
 
     def _feed_buffers(self):
         '''
-        Beckground thread to feed reassembled buffers to the parser
+        Background thread to feed reassembled buffers to the parser
         '''
         save = None
         while True:
@@ -564,7 +572,7 @@ class IOThread(threading.Thread):
             cmd = self.parse_control(data)
             if cmd['cmd'] == IPRCMD_ROUTE:
                 # routing request
-                family = cmd.get_attr('CTRL_ATTR_FAMILY_ID')[0]
+                family = cmd.get_attr('CTRL_ATTR_FAMILY_ID')
                 if family in self.families:
                     send = self.families[family]
                     self.rtable[sock] = send
@@ -576,7 +584,7 @@ class IOThread(threading.Thread):
                 # * ...
             elif cmd['cmd'] == IPRCMD_REGISTER:
                 # auth request
-                secret = cmd.get_attr('IPR_ATTR_SECRET')[0]
+                secret = cmd.get_attr('IPR_ATTR_SECRET')
                 if secret == self.secret:
                     self.controls.add(sock)
                     rsp['cmd'] = IPRCMD_ACK
@@ -638,6 +646,14 @@ class IOThread(threading.Thread):
                 # in this case add just an empty string
                 msg['header']['host'] = ''
 
+            # 8<--------------------------------------------------------------
+            # message filtering
+            # right now it is simply iterating callback list
+            for cr in self.callbacks:
+                if cr[0](msg):
+                    cr[1](msg, *cr[2])
+
+            # 8<--------------------------------------------------------------
             if key not in self.listeners:
                 key = 0
             if self.mirror and (key != 0) and (msg.raw is not None):
@@ -987,6 +1003,55 @@ class Netlink(object):
         else:
             del self.listeners[0]
 
+    def register_callback(self, callback, predicate=lambda x: True, args=None):
+        '''
+        Register a callback to run on a message arrival.
+
+        Callback is the function that will be called with the
+        message as the first argument. Predicate is the optional
+        callable object, that returns True or False. Upon True,
+        the callback will be called. Upon False it will not.
+        Args is a list or tuple of arguments.
+
+        Simplest example, assume ipr is the IPRoute() instance::
+
+            # create a simplest callback that will print messages
+            def cb(msg):
+                print(msg)
+
+            # register callback for any message:
+            ipr.register_callback(cb)
+
+        More complex example, with filtering::
+
+            # Set object's attribute after the message key
+            def cb(msg, obj):
+                obj.some_attr = msg["some key"]
+
+            # Register the callback only for the loopback device, index 1:
+            ipr.register_callback(cb,
+                                  lambda x: x.get('index', None) == 1,
+                                  (self, ))
+
+        Please note: you do **not** need to register the default 0 queue
+        to invoke callbacks on broadcast messages. Callbacks are
+        iterated **before** messages get enqueued.
+        '''
+        if args is None:
+            args = []
+        self.iothread.callbacks.append((predicate, callback, args))
+
+    def unregister_callback(self, callback):
+        '''
+        Remove the first reference to the function from the callback
+        register
+        '''
+        cb = tuple(self.iothread.callbacks)
+        for cr in cb:
+            if cr[1] == callback:
+                self.iothread.callbacks.pop(cb.index(cr))
+                return
+
     def _remove_queue(self, key):
         '''
         Flush the queue to the default one and remove it
